"""This module contains nodes for signal filtering."""

import numpy as np
import pandas as pd
from scipy import signal

from timeflux.core.branch import Branch
from timeflux.core.node import Node
from timeflux.nodes.window import TimeWindow
from timeflux_dsp.utils.filters import (
    construct_fir_filter,
    construct_iir_filter,
    design_edges,
)
from timeflux_dsp.utils.import_helpers import make_object


class DropRows(Node):
    """Decimate signal by an integer factor.

    This node uses Pandas computationally efficient functions to drop rows.
    By default, it simply transfers one row out of ``factor`` and drops the others.
    If ``method`` is `mean` (resp. median), it applies a rolling window of length
    equals ``factor``, computes the mean and returns one value per window.
    It maintains an internal state to ensure that every k'th sample is picked
    even across chunk boundaries.

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame.

    Args:
        factor (int): Decimation factor. Only every k'th sample will be
        transferred into the output.
        method (str|None): Method to use to drop rows.
                           If `None`, the values are transferred as it.
                           If `mean` (resp. `median`), the mean (resp. median)
                           of the samples is taken.

    Example:
       .. literalinclude:: /../examples/droprows.yaml
           :language: yaml

    Example:
        In this exemple, we generate white noise to stream and we drop one sample out
        of two using DropRows, setting:

        * ``factor`` = `2`
        * ``method`` = `None` (see orange trace) | ``method`` = `"mean"` (see green trace)

        .. image:: /static/image/droprows_io.svg
           :align: center


    Notes:
        Note that this node is not supposed to dejitter the timestamps, so if
        the input chunk is not uniformly sampled, the output chunk won’t be either.

        Also, this filter does not implement any anti-aliasing filter.
        Hence, it is recommended to precede this node by a low-pass filter
        (e.g., FIR or IIR) which cuts out below half of the new sampling rate.

    """

    def __init__(self, factor, method=None):

        super().__init__()
        self._factor = factor
        self._method = method
        self._previous = pd.DataFrame()

    def update(self):

        # copy the meta
        self.o.meta = self.i.meta

        # if nominal rate is specified in the meta, update it.
        if "rate" in self.o.meta:
            self.o.meta["rate"] /= self._factor

        # When we have not received data, there is nothing to do
        if not self.i.ready():
            return

        # At this point, we are sure that we have some data to process
        self.i.data = pd.concat([self._previous, self.i.data], axis=0, sort=True)

        n = self.i.data.shape[0]
        remaining = n % self._factor
        self.i.data, self._previous = np.split(self.i.data, [n - remaining])

        if self._method is None:
            # take every kth sample with k=factor starting from the k-1 position
            self.o.data = self.i.data.iloc[self._factor - 1 :: self._factor]
        else:
            # estimate rolling mean (or median) with window length=factor and take
            # every kth sample with k=factor starting from the k-1 position
            if self._method == "mean":
                self.o.data = (
                    self.i.data.rolling(
                        window=self._factor, min_periods=self._factor, center=False
                    )
                    .mean()
                    .iloc[self._factor - 1 :: self._factor]
                )
            elif self._method == "median":
                self.o.data = (
                    self.i.data.rolling(
                        window=self._factor, min_periods=self._factor, center=False
                    )
                    .median()
                    .iloc[self._factor - 1 :: self._factor]
                )


class Resample(Node):
    """Resample signal.

    This node calls the `scipy.signal.resample` function to decimate the signal
    using Fourier method.

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame.

    Args:
        factor (int): Decimation factor. Only every k'th sample will be
                      transferred into the output.
        window (str|list|float): Specifies the window applied to the signal
                      in the Fourier domain. Default: `None`.

    Example:
        .. literalinclude:: /../examples/resample.yaml
           :language: yaml

    Notes:
        This node should be used after a buffer to assure that the FFT window
        has always the same length.

    References:

        * `scipy.signal.resample <https://docs.scipy.org/doc/scipy-0.16.0/reference/generated/scipy.signal.resample.html>`_

    """

    def __init__(self, factor, window=None):

        super().__init__()
        self._factor = factor
        self._window = window
        self._previous = pd.DataFrame()

    def update(self):

        # copy the meta
        self.o.meta = self.i.meta

        # if nominal rate is specified in the meta, update it.
        if "rate" in self.o.meta:
            self.o.meta["rate"] /= self._factor

        # When we have not received data, there is nothing to do
        if not self.i.ready():
            return

        # At this point, we are sure that we have some data to process
        n = self.i.data.shape[0]

        if not self._previous.empty:
            self.i.data = pd.concat([self._previous, self.i.data], axis=0)

        if self.i.data.shape[0] % self._factor == 0:
            self._previous = pd.DataFrame()
        else:
            self._previous = self.i.data.iloc[(n // self._factor) * self._factor :]
            self.i.data = self.i.data.iloc[: (n // self._factor) * self._factor]

        self.o.data = pd.DataFrame(
            data=signal.resample(
                x=self.i.data.values, num=n // self._factor, window=self._window
            ),
            index=self.i.data.index[np.arange(0, n - 1, self._factor)],
            columns=self.i.data.columns,
        )


class IIRFilter(Node):
    """Apply IIR filter to signal.

    If ``sos`` is `None`, this node uses adapted methods from mne.filters to
    design the filter coefficients based on the specified parameters.
    If no transition band is given, default is to use :

    * l_trans_bandwidth =  min(max(l_freq * 0.25, 2), l_freq)
    * h_trans_bandwidth =   min(max(h_freq * 0.25, 2.), rate / 2. - h_freq)

    Else, it uses ``sos`` as filter coefficients.

    Once the kernel has been estimated, the node applies the filtering to each
    columns in ``columns`` using `scipy.signal.sosfilt` to generate the output given the input,
    hence ensures continuity  across chunk boundaries,

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame.

    Args:
        rate (float): Nominal sampling rate of the input data. If None, rate is get
                        from the meta.
        order (int, optional): Filter order. Default: `None`.
        frequencies (list|None): Transition frequencies. Ignored when sos is given.
        filter_type (str|None): Filter mode (`lowpass`, `highpass`, `bandstop`, `bandpass`).
                        Default: `bandpass`. Ignored when sos is given.
        sos (array|None, optional) : Array of second-order sections (sos) representation,
                                    must have shape (n_sections, 6). Default: `None`.
        kwargs: keyword arguments to pass to the filter constructor


    Example:
        In this example, we generate a signal that is the sum of two sinus with
        respective periods of 1kHz and 15kHz and respective amplitudes of 1 and 0.5.
        We stream this signal using the IIRFilter node, designed for lowpass
        filtering at cutoff frequency 6kHz, order 3.

        * ``order`` = `3`
        * ``freqs`` = `[6000]`
        * ``mode`` = `'lowpass'`

        We plot the input signal, the output signal and the corresponding offline filtering.

        .. image:: /static/image/iirfilter_io.svg
           :align: center

    Notes:
        This node ensures continuity across chunk boundaries, using a recursive algorithm,
        based on a cascade of biquads filters.

        The filter is initialized to have a minimal step response, but needs a
        'warm up' period for the filtering to be stable, leading to small artifacts
        on the first few chunks.

        The IIR filter is faster than the FIR filter and delays the signal less
        but this delay is not constant and the stability not  ensured.

    References:

            * `Real-Time IIR Digital Filters <http://www.eas.uccs.edu/~mwickert/ece5655/lecture_notes/ece5655_chap8.pdf>`_
            * `scipy.signal.sosfilt <https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.sosfilt.html>`_

    """

    def __init__(
        self, frequencies=None, rate=None, filter_type="bandpass", sos=None, **kwargs
    ):

        super().__init__()
        # self._order = order
        self._frequencies = frequencies
        self._filter_type = filter_type
        self._kwargs = dict(order=None, design="butter", pass_loss=3.0, stop_atten=50.0)
        self._kwargs.update(kwargs)
        self._rate = rate
        self._zi = None
        self._sos = None
        self._sos_custom = sos
        self._columns = None

    def update(self):

        # copy the meta
        self.o = self.i

        # When we have not received data, there is nothing to do
        if not self.i.ready():
            return

        # At this point, we are sure that we have some data to process
        if self._columns is None:
            self._columns = self.i.data.columns

        # set rate from the data if it is not yet given
        if self._rate is None:
            self._rate = self.i.meta.get("rate", None)
            if self._rate is None:
                # If there is no rate in the meta, set rate to 1.0
                self._rate = 1.0
                self.logger.warning(
                    f"Nominal rate not supplied, considering " f"1.0 Hz instead. "
                )
            else:
                self.logger.info(f"Nominal rate set to {self._rate}. ")

        if self._sos is None:
            self._design_sos()
        if self._zi is None:
            zi0 = signal.sosfilt_zi(self._sos)
            self._zi = np.stack(
                [
                    (zi0 * self.i.data.iloc[0, k_col])
                    for k_col in range(len(self._columns))
                ],
                axis=1,
            )
        port_o, self._zi = signal.sosfilt(self._sos, self.i.data.values.T, zi=self._zi)
        self.o.data = pd.DataFrame(
            port_o.T, columns=self._columns, index=self.i.data.index
        )

    def _design_sos(self):

        if self._sos_custom is None:
            # Calculate an IIR filter kernel for a given sampling rate.
            self._sos, self._freqs = construct_iir_filter(
                rate=self._rate,
                frequencies=self._frequencies,
                filter_type=self._filter_type,
                output="sos",
                **self._kwargs,
            )
        else:
            if self._sos_custom.shape[1] == 6:
                self._sos = self._sos_custom
            else:
                raise ValueError(
                    f"sos must have shape (n_sections, 6), received {self._sos_custom.shape} instead. "
                )


class IIRLineFilter(Node):
    """Apply multiple Notch IIR Filter in series.

        Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame.

    Args:
        rate (float): Nominal sampling rate of the input data. If None, rate is get
                  from the meta.
        edges_center: List with center of the filters.
        orders (tuple|int|None): List with orders of the filters.
                            If int, the same order will be used for all filters.
                            If None, order 2 will be used for all filters.
        edges_width: List with orders of the filters.
                     If int, the same order will be used for all filters.
                     If None, width of 3 (Hz) will be used for all filters.

    """

    def __init__(
        self,
        rate=None,
        edges_center=(50, 60, 100, 120),
        orders=(2, 1, 1, 1),
        edges_width=(3, 3, 3, 3),
    ):

        super().__init__()

        orders = orders or 2
        edges_width = edges_width or 3

        if isinstance(orders, int):
            orders = [orders] * len(edges_center)

        if isinstance(edges_width, int):
            edges_width = [edges_width] * len(edges_center)

        filter_type = "bandstop"
        self._nodes = []
        for edge_center, edge_width, order in zip(edges_center, edges_width, orders):
            frequencies = [edge_center - edge_width, edge_center + edge_width]
            self._nodes.append(
                IIRFilter(
                    rate=rate,
                    order=order,
                    frequencies=frequencies,
                    filter_type=filter_type,
                )
            )

    def update(self):

        # When we have not received data, there is nothing to do
        if not self.i.ready():
            return
        # At this point, we are sure that we have some data to process
        # initialize output port
        self.o = self.i
        # apply each filter in series
        for node in self._nodes:
            node.i.data = self.o.data
            node.update()
            self.o.data = node.o.data


class FIRFilter(Node):
    """Apply FIR filter to signal.

    If ``coeffs`` is `None`, this node uses adapted methods from *mne.filters*
    to design the filter coefficients based on the specified parameters.
    If no transition band is given, default is to use:

    * l_trans_bandwidth =  min(max(l_freq * 0.25, 2), l_freq)
    * h_trans_bandwidth =   min(max(h_freq * 0.25, 2.), fs / 2. - h_freq)

    Else, it uses ``coeffs`` as filter coefficients.

    It applies the filtering to each columns in ``columns`` using `scipy.signal.lfilter`
    to generate the output given the input,
    hence ensures continuity  across chunk boundaries,

    The delay introduced is estimated and stored in the meta ``FIRFilter``, ``delay``.

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame and meta.

    Args:
        rate (float): Nominal sampling rate of the input data. If None, rate is get
                      from the meta.
        columns (list|'all', optional): Columns to apply filter on. Default: `all`.
        order (int): Filter order.
        frequencies (list): Transition frequencies.
        filter_type (str, optional): Filter mode (`lowpass`, `highpass`, `bandstop`
                                    or `bandpass`). Default: `bandpass`.
        coeffs (array|None, optional): Custom coeffs to pass as ``b`` in `signal.filter`.
                                        Default: `None`.
        kwargs: keyword arguments to pass to the filter constructor (window, phase,... )

    Example:
        In this exemple, we generate a signal that is the sum of two sinus with
        respective periods of 1kHz and 15kHz and respective amplitudes of 1 and 0.5.
        We stream this signal using the FIRFilter node, designed for lowpass
        filtering at cutoff frequency 6kHz, order 20.

        * ``order`` = `20`
        * ``freqs`` = `[6000, 6100]`
        * ``mode`` = `'lowpass'`

        The FIR is a linear phase filter, so it allows one to correct for the
        introduced delay. Here, we retrieve the input sinus of period 1kHz.
        We plot the input signal, the output signal, the corresponding offline
        filtering and the output signal after delay correction.

        .. image:: /static/image/firfilter_io.png
           :align: center

    Notes:
        The FIR filter ensures a linear phase response, but is computationnaly
        more costly than the IIR filter.

        The filter is initialized to have a minimal step response, but needs a
        'warmup' period for the filtering to be stable, leeding to small artifacts on
        the first few chunks.

    """

    def __init__(
        self,
        frequencies,
        rate=None,
        columns="all",
        order=20,
        filter_type="bandpass",
        coeffs=None,
        **kwargs,
    ):

        super().__init__()
        self._order = order
        self._frequencies = frequencies
        self._mode = filter_type
        self._rate = rate
        self._columns = columns if columns != "all" else None
        self._coeffs_custom = coeffs
        self._kwargs = dict(design="firwin2", phase="linear", window="hamming")
        self._kwargs.update(kwargs)

        # Initialize the filter kernels and states, one per stream
        self._zi = {}  # FIR filter states, one per column
        self._coeffs = None  # FIR filter coeffs
        self._delay = None  # FIR filter delays

    def update(self):

        # copy the meta
        self.o = self.i

        # When we have not received data, there is nothing to do
        if not self.i.ready():
            return

        # At this point, we are sure that we have some data to process
        if self._columns is None:
            self._columns = self.i.data.columns

        # set rate from the data if it is not yet given
        if self._rate is None:
            self._rate = self.i.meta.get("rate", None)
            if self._rate is None:
                # If there is no rate in the meta, set rate to 1.0
                self._rate = 1.0
                self.logger.warning(
                    f"Nominal rate not supplied, considering " f"1.0 Hz instead. "
                )
            else:
                self.logger.info(f"Nominal rate set to {self._rate}. ")

        self._coeffs, self._delay = self._design_filter()

        for column in self._columns:
            if column not in self._zi:
                zi0 = signal.lfilter_zi(self._coeffs, 1.0)
                self._zi[column] = zi0 * self.i.data[column].values[0]
            port_o_col, self._zi[column] = signal.lfilter(
                b=self._coeffs,
                a=1.0,
                x=self.i.data[column].values.T,
                zi=self._zi[column],
            )
            # self.o.meta.update({'FIRFilter': {'delay': self._delay}})
            self.o.data.loc[:, column] = port_o_col
            # update delay
            delay = self.o.meta.get("delay") or 0.0
            delay += self._delay
            self.o.meta.update({"delay": delay})

    def _design_filter(self):
        """Calculate an FIR filter kernel for a given sampling rate."""

        nyq = self._rate / 2.0

        if self._coeffs_custom is None:
            edges, gains, _, _ = design_edges(
                frequencies=self._frequencies, nyq=nyq, mode=self._mode
            )
            fir_coeffs = construct_fir_filter(
                self._rate, edges, gains, self._order, **self._kwargs
            )
        else:
            fir_coeffs = self._coeffs_custom
        warmup = self._order - 1
        fir_delay = (warmup / 2) / self._rate
        return fir_coeffs, fir_delay


class Scaler(Node):
    """Apply a sklearn scaler

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame and meta.

    Args:
        method (str): Name of the scaler object (see https://scikit-learn.org/stable/modules/classes.html#module-sklearn.preprocessing)
        **kwargs: keyword arguments  to initialize the scaler.

    """

    def __init__(self, method="sklearn.preprocessing.StandardScaler", **kwargs):

        super().__init__()
        try:
            self._scaler = make_object(method, kwargs)
        except AttributeError:
            raise ValueError("Cannot make object from {method}".format(method=method))

    def update(self):

        if not self.i.ready():
            return

        # scale the signal
        self.o.data = pd.DataFrame(
            data=self._scaler.fit_transform(self.i.data.values),
            columns=self.i.data.columns,
        )
        if len(self.o.data) == len(self.i.data):
            self.o.data.index = self.i.data.index


class AdaptiveScaler(TimeWindow):
    """Scale the data adaptively.
    This nodes transforms the data using a sklearn scaler object that is continuously fitted on a rolling window.

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame and meta.

    Args:
       length (float): The length of the window, in seconds.
       method (str): Name of the scaler object (see https://scikit-learn.org/stable/modules/classes.html#module-sklearn.preprocessing)
       dropna (bool): Whether or not NaN should be dropped before fitting the estimator. Default to False.
       **kwargs : keyword arguments  to initialize the scaler.
    """

    def __init__(
        self,
        length,
        method="sklearn.preprocessing.StandardScaler",
        dropna=False,
        **kwargs,
    ):

        super().__init__(length=length, step=0)
        self._fitted = False
        self._dropna = dropna
        try:
            self._scaler = make_object(method, kwargs)
        except AttributeError:
            raise ValueError(f"Module sklearn.preprocessing has no object {method}")

    def update(self):

        if not self.i.ready():
            return

        # At this point, we are sure that we have some data to process
        super().update()

        # if the window output is ready, fit the scaler with its values
        if self.o.ready():
            x = self.o.data.values
            if self._dropna:
                x = x[~np.isnan(x)]
            self._scaler.fit(x)
            self._fitted = True
            self.o.clear()

        # copy meta
        self.o.meta = self.i.meta
        # if the scaler has been fitted, transform the current data
        if self._fitted and self.i.ready():
            transformed_data = self._scaler.transform(self.i.data.values)
            self.o.data = pd.DataFrame(
                data=transformed_data,
                columns=self.i.data.columns,
                index=self.i.data.index,
            )


class FilterBank(Branch):
    """Apply multiple IIR Filters to the signal and stack the components horizontally

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame.

    Args:
        rate (float): Nominal sampling rate of the input data. If None, rate is get
            from the meta.
        filters (dict|None): Define the iir filter to apply given its name and its params.
    """

    def __init__(self, filters, method="IIRFilter", rate=None, **kwargs):
        super().__init__()
        self._filters = filters

        graph = {"nodes": [], "edges": []}
        graph["nodes"].append(
            {
                "id": "stack",
                "module": "timeflux_dsp.nodes.helpers",
                "class": "Concat",
            }
        )

        for filter_name, filter_params in self._filters.items():
            filter_params.update({"rate": rate})
            filter_params.update(kwargs)
            iir = {
                "id": filter_name,
                "module": "timeflux_dsp.nodes.filters",
                "class": method,
                "params": filter_params,
            }
            rename_columns = {
                "id": f"rename_{filter_name}",
                "module": "timeflux.nodes.axis",
                "class": "AddSuffix",
                "params": {"suffix": f"_{filter_name}"},
            }
            graph["nodes"] += [iir, rename_columns]
            graph["edges"] += [
                {"source": filter_name, "target": f"rename_{filter_name}"},
                {"source": f"rename_{filter_name}", "target": f"stack:{filter_name}"},
            ]

        self.load(graph)

    def update(self):
        # When we have not received data, there is nothing to do
        if not self.i.ready():
            return
        # set the data in input of each filter
        for filter_name in self._filters.keys():
            self.set_port(filter_name, port_id="i", data=self.i.data, meta=self.i.meta)

        self.run()

        self.o = self.get_port("stack", port_id="o")



from collections import deque

import numpy as np
import pandas as pd
from timeflux.core.node import Node
from timeflux.helpers.clock import now


class LocalDetect(Node):
    """Detect peaks and valleys in live 1D signal

    This node uses a simple algorithm to detect peaks in real time.
    When a local extrema (peak or valley) is detected, an event is sent with the nature
    specified in the label column (``peak``/ ``valley``) andcthe characteristics in the
    data column, giving:

    - **value**: Amplitude of the extrema.
    - **lag**: Time laps between the extrema and its detection.
    - **interval**: Duration between two extrema os same nature.

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Events output, provides DataFrame.

    Args:
        delta (float): Threshold for peak/valley matching in amplitude. This can be seen as the minimum significant
                    change enough to detect a peak/valley.
        tol (float): Tolerence for peak/valley matching, in seconds. This can be seen as the minimum time difference
                    between two peaks/valleys.
        reset (float): Reset threshold, in seconds.
                    This can be seen as the maximum duration of plausible transitions between peaks and valleys.
                    Default: None.

    Example:
       .. literalinclude:: /../examples/droprows.yaml
           :language: yaml

    Example:
        In this example, we stream a photoplethysmogram signal scaled between -1 and 1 and we use the node
        RealTimeDetect to detect peaks and valleys.

        * ``delta`` = `0.1`
        * ``tol`` = `0.5`
        * ``reset`` = None

        .. image:: /static/image/realtimepeaks_io.svg
           :align: center

        self.o.data::

                                            label                                               data
            2018-11-19 11:06:39.620900000    peak  {'value': [1.0054607391357422], 'lag': 0.03125, 'interval': 0.654236268}
            2018-11-19 11:06:39.794709043  valley  {'value': [-1.0110111236572266], 'lag': 0.046875, 'interval': 0.654236268}
            2018-11-19 11:06:40.605209027    peak  {'value': [0.9566234350204468],  'lag': 0.033197539, 'interval': 0.984309027}
            2018-11-19 11:06:40.761455675  valley  {'value': [-1.0549353361129759], 'lag': 0.048816963, 'interval': 0.810499984}


    Notes:

        This peak detection is considered real-time since it does not require buffering the data. However, the detection
        method necessarily involve a lag between the actual peak and its detection.
        Indeed, the internal state of the node is either "looking for a peak" or "looking for a valley".

        - If the node is "looking for a peak", it means it computes local maxima
            until the signal drops significantly (ie. more than ``delta``)
        - If the node is "looking for a valley", it means it computes local minima
            until the signal rises significantly (ie. more than ``delta``)

        The "last local extrema" is set to a peak (resp. valley) as soon as the signal drops (resp. rises) significantly.
        Hence, there is an intrinsic lag in the detection, that is directly linked to parameter ``delta``.

        Hence, by decreasing delta, we minimize the lag. But if delta is too small,
        we'll suffer from false positive detection, unless ``tol`` is tuned to avoid too closed detections.
        The parameters should be tuned depending on the nature of the data, ie. their dynamic, quality, shapes.

        See the illustration above:

        .. image:: /static/image/realtimepeaks_illustration.png
            :align: center


    References:

        * `Matlab function <http://billauer.co.il/peakdet.html>`_
        * `Publication <https://www.ncbi.nlm.nih.gov/pubmed/27027672>`_


    **Todo**

        * allow for adaptive parametrization.

    """

    def __init__(self, delta, tol, reset=None):

        super().__init__()
        self._delta = delta  # Peak threshold
        self._tol = tol  # Tolerence for peak matching, in seconds. This can be seen as the minimum time difference
        self._reset_states()
        self._last = pd.to_datetime(now())  # Last timestamp
        self._reset = reset

    def _reset_states(self):

        """Reset peak detection internal state."""
        self._mxv = -np.Inf  # local max value
        self._mnv = np.Inf  # local min value
        self._mxt = None  # local max time
        self._mnt = None  # local min time
        self._lfm = True  # look for max

        self._last_peak = None
        self._last_valley = None

    def update(self):

        # copy the meta
        self.o.meta = self.i.meta

        # When we have not received data, there is nothing to do
        if self.i.data is None or self.i.data.empty:
            return

        if self.i.data.shape[1] != 1:
            self.logger.warning(
                f"Peak detection expects data with one column, received "
                f"{self.i.data.shape[1]}. Considering the first one. "
            )
            self.i.data = self.i.data.take([0], axis=1)

        column_name = self.i.data.columns[0]
        # At this point, we are sure that we have some data to process
        self.o.data = pd.DataFrame()

        for (value, timestamp) in zip(self.i.data.values, self.i.data.index):
            if self._reset is not None:
                if (self._last - timestamp).total_seconds() > self._reset:
                    self._reset_states()
            self._last = timestamp
            # Peak detection
            detected = self._on_sample(value=value, timestamp=timestamp)
            # Append event
            if detected:
                self.o.data = self.o.data.append(
                    pd.DataFrame(
                        index=[self.i.data.index[-1]],  # detected[0]
                        data=np.array(
                            [
                                [detected[1]],
                                [
                                    {
                                        "value": detected[2][0],
                                        "lag": detected[3],
                                        "interval": detected[4],
                                        "column_name": column_name,
                                        "detection_time": str(self.i.data.index[-1]),
                                        "now": str(now()),
                                        "extremum_time": str(detected[0]),
                                    }
                                ],
                            ]
                        ).T,
                        columns=["label", "data"],
                    )
                )
                self.o.meta = {"column_name": column_name}

    def _on_sample(self, value, timestamp):
        """Peak detection"""

        if self._last_peak is None:
            self._last_peak = timestamp
        if self._last_valley is None:
            self._last_valley = timestamp

        if value > self._mxv:
            self._mxv = value
            self._mxt = timestamp

        if value < self._mnv:
            self._mnv = value
            self._mnt = timestamp
        if self._lfm:
            if value < self._mxv - self._delta:
                _interval = (self._mxt - self._last_peak).total_seconds()
                self._mnv = value
                self._mnt = timestamp
                self._lfm = False
                if (self._mxt - self._last_peak).total_seconds() > self._tol:
                    self._last_peak = self._mxt
                    return (
                        self._mxt,
                        "peak",
                        self._mxv,
                        (timestamp - self._mxt).total_seconds(),
                        _interval,
                    )
        else:
            if value > self._mnv + self._delta:
                _interval = (self._mnt - self._last_valley).total_seconds()
                self._mxv = value
                self._mxt = timestamp
                self._lfm = True
                if (self._mnt - self._last_valley).total_seconds() > self._tol:
                    self._last_valley = self._mnt
                    return (
                        self._mnt,
                        "valley",
                        self._mnv,
                        (timestamp - self._mnt).total_seconds(),
                        _interval,
                    )
        return False


class RollingDetect(Node):
    """Detect peaks and valleys on a rolling window of analysis in  1D signal
    This node uses a buffer to compute local extrema and detect peaks in real time.
    When a local extrema (peak or valley) is detected, an event is sent with the
    nature specified in the label column ("peak"/"valley) and
    the characteristics in the data column, giving:

    - **value**: Amplitude of the extrema.
    - **lag**: Time laps between the extrema and its detection.
    - **interval**: Duration between two extrema os same nature.

    Args:
        window (float): Window of analysis in seconds, on which local max/min is computed.
        tol (float): Tolerance for peak/valley matching, in seconds.
        This can be seen as the minimum time difference
        between two peaks/valleys.

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Events output, provides DataFrame.
    """

    def __init__(
        self, length: object = 0.5, tol: object = 0.1, rate: object = None
    ) -> object:

        super().__init__()
        self._length = length  # Window of analysis
        self._tol = tol  # Tolerance for peak matching, in seconds.
        self._rate = rate  # Signal rate. TODO: this should be in the meta
        self._n = int(self._rate * self._length)
        # (this can be seen as the minimum time difference between two peaks)

        self._column = None
        self._reset_states()

    def _reset_states(self):
        """Reset peak detection internal state."""

        self._values_buffer = deque(maxlen=2 * self._n)
        self._timestamps_buffer = deque(maxlen=2 * self._n)

        self._last_peak = None
        self._last_valley = None

        # self._peak_interval = timedelta(seconds=1)
        # self._valley_interval = timedelta(seconds=1)

        self._ready = False

    def update(self):

        # copy the meta
        self.o.meta = self.i.meta

        # When we have not received data, there is nothing to do
        if not self.i.ready():
            return
        # # At this point, we are sure that we have some data to process
        if self.i.data.shape[1] != 1:
            self.logger.warning(
                f"Peak detection expects data with one column, received "
                f"{self.i.data.shape[1]}. Considering the first one. "
            )
            self.i.data = self.i.data.take([0], axis=1)

        self._last = self.i.data.index[-1]
        if not self._ready:
            self._column = self.i.data.columns[0]
            # if self._last_peak is None:
            self._last_peak = self._last_valley = self.i.data.index[0]
            self._values_buffer += [0] * 2 * self._n
            self._timestamps_buffer += [self.i.data.index[0]] * 2 * self._n
            self._ready = True
        self.o.meta = {"column_name": self._column}

        self.o.data = pd.DataFrame()

        for (value, timestamp) in zip(self.i.data.values, self.i.data.index):
            # Peak detection
            detected = self._on_sample(value=value, timestamp=timestamp)
            if detected:
                self.o.data = self.o.data.append(
                    pd.DataFrame(
                        index=[detected[0]],
                        data=np.array(
                            [
                                [detected[1]],
                                [
                                    {
                                        "value": detected[2],
                                        "lag": detected[3],
                                        "interval": detected[4],
                                        "column_name": self._column,
                                        "detection_time": str(self._last),
                                        "now": str(now()),
                                        "extremum_time": str(detected[0]),
                                    }
                                ],
                            ]
                        ).T,
                        columns=["label", "data"],
                    )
                )
                self.o.meta = {"column_name": self._column}

    def _on_sample(self, value, timestamp):
        """Peak detection"""
        self._values_buffer.append(value)
        self._timestamps_buffer.append(timestamp)

        if self._values_buffer[self._n] == max(self._values_buffer):
            peak = self._timestamps_buffer[self._n]
            # peak candidate
            _interval = (peak - self._last_peak).total_seconds()
            if _interval > self._tol:
                # peak detected
                _lag = (self._last - peak).total_seconds()
                self._last_peak = peak
                return peak, "peak", self._values_buffer[self._n], _lag, _interval
        elif self._values_buffer[self._n] == min(self._values_buffer):
            valley = self._timestamps_buffer[self._n]
            # peak candidate
            _interval = (valley - self._last_valley).total_seconds()
            if _interval > self._tol:
                # peak detected
                _lag = (self._last - valley).total_seconds()
                self._last_valley = valley
                return valley, "valley", self._values_buffer[self._n], _lag, _interval
        return False


class Rate(Node):
    """Computes rate of an event given its label.

    This node computes the inverse duration (ie. instantaneous rate) between onsets of successive events with a marker
    matching the ``event_trigger`` in the ``event_label`` column of the event input,

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame.

    Args:
        event_trigger (string): The marker name.
        event_label (string): The column to match for event_trigger.
    """

    def __init__(self, event_trigger="peak", event_label="label"):

        super().__init__()
        self._event_trigger = event_trigger
        self._event_label = event_label
        self._column_name = None
        self._last = None

    def _reset_states(self):
        """Reset peak detection internal state."""

        self._last = None

    def update(self):

        # copy the meta
        self.o.meta = self.i.meta

        # When we have not received data, there is nothing to do
        if not self.i.ready():
            return

        # At this point, we are sure that we have some data to process
        self.o.data = None

        target_index = self.i.data[
            self.i.data[self._event_label] == self._event_trigger
        ].index

        if self._column_name is None and len(self.i.meta) > 0:
            self._column_name = self.i.meta["column_name"]

        if not target_index.empty:

            if self._last is None:
                self._last = target_index[0]

            target_index = [self._last] + list(target_index)

            rate_values = [
                (np.timedelta64(1, "s") / a)
                if (a != 0 * np.timedelta64(1, "s"))
                else None
                for a in list(np.diff(target_index))
            ]

            self.o.data = pd.DataFrame(
                index=target_index[1:], data=rate_values, columns=[self._column_name]
            )
            self._last = target_index[-1]



import numpy as np
from timeflux.core.node import Node


class Discretize(Node):
    """Discretize data based on defined ranges
    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame.

    Args:
        range (dict): dictionary with keys corresponding to the discrete labels and values
                      are lists of tuple with boundaries.
        default (float|str|None):
    """

    def __init__(self, range, default=None):
        super().__init__()

        if default is None:
            default = np.NaN
        default = default
        self._default = default
        self._range = range
        for k, range_values in self._range.items():
            for v in range_values:
                if v[0] is None:
                    v[0] = -np.inf
                if v[1] is None:
                    v[1] = np.inf

    def _discretize(self, x):
        for k, range_values in self._range.items():
            for v in range_values:
                if (x >= v[0]) & (x <= v[1]):
                    return k
        return self._default

    def update(self):
        if self.i.ready():
            self.o = self.i
            if self.o.data is not None:
                self.o.data = self.o.data.applymap(lambda x: self._discretize(x))



"""Tests for nodes from timeflux_dsp.nodes.spectral"""

import numpy as np
import pandas as pd
import pytest
import xarray as xr
from timeflux.helpers.testing import DummyData

from timeflux_dsp.nodes.spectral import FFT

fs = 10

data = DummyData(rate=fs, jitter=0.05)
all_data = data.next(50)


def test_welch():
    data.reset()

    node = FFT(fs=fs, return_onesided=False)
    node.i.data = data.next(5)

    node.update()
    expected_freqs = [0.0, 2.0, 4.0, -4.0, -2.0]
    expected_times = [pd.Timestamp("2018-01-01 00:00:00.396560186")]
    expected_data = np.array(
        [
            [
                2.687793 + 0.0j,
                2.69977 + 0.0j,
                4.158542 + 0.0j,
                2.907866 + 0.0j,
                2.979773 + 0.0j,
            ],
            [
                -0.32328042 + 0.45056971j,
                -0.09741619 - 0.84999621j,
                0.19777914 - 0.14955481j,
                0.33690762 + 1.2010184j,
                0.65131083 + 0.03780588j,
            ],
            [
                -0.55778358 - 0.64687062j,
                0.10228369 + 0.5354582j,
                -0.09468514 + 0.40190712j,
                0.03972188 - 0.40916112j,
                -0.12479483 + 0.90610597j,
            ],
            [
                -0.55778358 + 0.64687062j,
                0.10228369 - 0.5354582j,
                -0.09468514 - 0.40190712j,
                0.03972188 + 0.40916112j,
                -0.12479483 - 0.90610597j,
            ],
            [
                -0.32328042 - 0.45056971j,
                -0.09741619 + 0.84999621j,
                0.19777914 + 0.14955481j,
                0.33690762 - 1.2010184j,
                0.65131083 - 0.03780588j,
            ],
        ]
    )

    # output for "MultiIndex" soon-deprecated version
    # expected = pd.DataFrame(index = pd.MultiIndex.from_product([expected_times, expected_freqs], names = ["times", "freqs"]), data=expected_data)

    expected = xr.DataArray(
        np.stack([expected_data], 0),
        coords=[expected_times, expected_freqs, data._data.columns],
        dims=["time", "freq", "space"],
    )

    xr.testing.assert_allclose(node.o.data, expected, rtol=1e-06)


def test_fft_invalid():
    data.reset()
    with pytest.raises(ValueError):
        node = FFT(nfft=5, fs=fs, return_onesided=False)
        node.i.data = data.next(100)
        node.update()


def test_fft_nfft():
    data.reset()
    node = FFT(nfft=None, fs=fs, return_onesided=False)
    node.i.data = data.next(10)
    node.update()
    assert node._nfft == 10



import numpy as np
import pandas as pd
from neurokit import ecg_process
from timeflux.core.branch import Branch
from timeflux.core.node import Node
from timeflux.nodes.window import Window


class Discretize(Node):
    """Discretize data based on amplitude range.

    Attributes:
    i (Port): Default input, expects DataFrame.
    o (Port): Default output, provides DataFrame.

    Args:
        range (dict): Dictionary with keys are discrete value and values
        are tuple with corresponding data ranges.
        default: Default discrete value (for data that are not contained in any range)
    """

    def __init__(self, range, default=None):

        super().__init__()
        self._default = default or np.NaN
        self._range = range
        for k, range_values in self._range.items():
            for v in range_values:
                if v[0] is None:
                    v[0] = -np.inf
                if v[1] is None:
                    v[1] = np.inf

    def _discretize_sqi(self, x):

        for k, range_values in self._range.items():
            for v in range_values:
                if (x >= v[0]) & (x <= v[1]):
                    return k
        return self._default

    def update(self):

        if not self.i.ready():
            return

        self.o = self.i
        self.o.data = self.o.data.applymap(lambda x: self._discretize_sqi(x))


class ECGQuality(Window):
    """Estimate ECG Quality

    This nodes estimates ECG Quality using neurokit toolbox, by applying function
    `ecg_process` on a rolling window.

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame.

    Args:
        rate (float): Nominal sampling rate of the input data. If None, rate is get
            from the meta.
        length (float): The length of the window, in seconds.
        step (float): The sliding step, in seconds.

    """

    def __init__(self, rate, length, step):

        try:
            from neurokit.bio.bio_ecg import ecg_process
            import sklearn.externals.joblib  # Fix for bug on neurokit
        except ModuleNotFoundError:
            self.logger.error("Neurokit is not installed")

        self._rate = rate
        super().__init__(length=length, step=step)

    def update(self):

        if not self.i.ready():
            return

        # set rate from the data if it is not yet given
        if self._rate is None:
            try:
                self._rate = self.i.meta.pop("nominal_rate")
                self.logger.info(f"Nominal rate set to {self._rate}. ")
            except KeyError:
                # If there is no rate in the meta, set rate to 1.0
                self._rate = 1.0
                self.logger.warning(
                    f"Nominal rate not supplied, considering " f"1.0 Hz instead. "
                )

        # At this point, we are sure that we have some data to process
        super().update()
        # if the window output is ready, fit the scaler with its values
        if self.o.ready() and not self.o.data.dropna().empty:
            X = self.o.data.dropna().values[:, 0]
            try:
                df = ecg_process(X, sampling_rate=self._rate, hrv_features=None)["df"]
                avg_rate = df["Heart_Rate"].dropna().mean()
                if avg_rate == np.NaN:
                    avg_quality = 0.0
                else:
                    avg_quality = df["ECG_Signal_Quality"].dropna().median()
            except Exception as e:
                self.logger.debug(e)
                avg_quality = 0.0
            self.o.data = pd.DataFrame(
                columns=["ecg_quality"],
                data=[avg_quality],
                index=[self.i.data.index[-1]],
            )


class LineQuality(Branch):
    """Estimate level of line noise

    This nodes estimates LineNoise as the ratio between good power and total power on a rolling
    window.
    Good power is defined as the sum of squared samples for signal after bandpass and Notch filtering.
    Total power is defined as the sum of squared samples for signal after bandpass filtering only.

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame.

    Args:
        rate (float): Nominal sampling rate of the input data. If None, rate is get
          from the meta.
       length (float): The length of the window, in seconds.
       step (float): The sliding step, in seconds.

    """

    def __init__(
        self,
        rate,
        range,
        window_length=3,
        window_step=0.5,
        bandpass_frequencies=(1, 65),
        line_centers=(50, 100, 150),
    ):

        super().__init__()
        graph = {
            "nodes": [
                {
                    "id": "linefilter",
                    "module": "timeflux_dsp.nodes.filters",
                    "class": "IIRLineFilter",
                    "params": {"rate": rate, "edges_center": line_centers},
                },
                {
                    "id": "bandpass",
                    "module": "timeflux_dsp.nodes.filters",
                    "class": "IIRFilter",
                    "params": {
                        "rate": rate,
                        "order": 3,
                        "frequencies": bandpass_frequencies,
                    },
                },
                {
                    "id": "square_good",
                    "module": "timeflux.nodes.apply",
                    "class": "ApplyMethod",
                    "params": {
                        "method": "numpy.square",
                        "apply_mode": "universal",
                    },
                },
                {
                    "id": "square_total",
                    "module": "timeflux.nodes.apply",
                    "class": "ApplyMethod",
                    "params": {
                        "method": "numpy.square",
                        "apply_mode": "universal",
                    },
                },
                {
                    "id": "window_good",
                    "module": "timeflux.nodes.window",
                    "class": "Window",
                    "params": {
                        "length": window_length,
                        "step": window_step,
                    },
                },
                {
                    "id": "window_total",
                    "module": "timeflux.nodes.window",
                    "class": "Window",
                    "params": {
                        "length": window_length,
                        "step": window_step,
                    },
                },
                {
                    "id": "sum_good",
                    "module": "timeflux.nodes.apply",
                    "class": "ApplyMethod",
                    "params": {
                        "method": "numpy.sum",
                        "apply_mode": "reduce",
                    },
                },
                {
                    "id": "sum_total",
                    "module": "timeflux.nodes.apply",
                    "class": "ApplyMethod",
                    "params": {
                        "module_name": "numpy.sum",
                        "apply_mode": "reduce",
                    },
                },
                {
                    "id": "divide",
                    "module": "timeflux.nodes.expression",
                    "class": "Expression",
                    "params": {
                        "expr": "i_1/i_2",
                        "eval_on": "ports",
                    },
                },
                {
                    "id": "log",
                    "module": "timeflux.nodes.apply",
                    "class": "ApplyMethod",
                    "params": {
                        "module_name": "numpy.log",
                        "apply_mode": "universal",
                    },
                },
                {
                    "id": "discretize",
                    "module": "timeflux_dsp.nodes.quality",
                    "class": "Discretize",
                    "params": {"range": range},
                },
            ],
            "edges": [
                {"source": "bandpass", "target": "linefilter"},
                {"source": "linefilter", "target": "square_good"},
                {"source": "bandpass", "target": "square_total"},
                {"source": "square_good", "target": "window_good"},
                {"source": "square_total", "target": "window_total"},
                {"source": "window_good", "target": "sum_good"},
                {"source": "window_total", "target": "sum_total"},
                {"source": "sum_good", "target": "divide:1"},
                {"source": "sum_total", "target": "divide:2"},
                {"source": "divide", "target": "log"},
                {"source": "log", "target": "discretize"},
            ],
        }
        self.load(graph)

    def update(self):

        # When we have not received data, there is nothing to do
        if not self.i.ready():
            return

        # copy the meta
        self.o.meta = self.i.meta

        self.set_port("bandpass", port_id="i", data=self.i.data, meta=self.i.meta)
        self.run()
        self.o = self.get_port("discretize", port_id="o")


class AmplitudeQuality(Branch):
    """Estimate discrete signal quality index based on a temporal feature from the amplitude.

    This nodes rolls a window and applies a numpy function  given by ``method``
    (eg. ptp, max, min, mean...) over rows and discretize the result based on ``range`` .

    Attributes:
        i (Port): Default input, expects DataFrame.
        o (Port): Default output, provides DataFrame.
    """

    def __init__(self, range, window_length=3, window_step=0.5, method="ptp"):

        super().__init__()
        self._range = range
        for k, range_values in self._range.items():
            for v in range_values:
                if v[0] is None:
                    v[0] = -np.inf
                if v[1] is None:
                    v[1] = np.inf
        graph = {
            "nodes": [
                {
                    "id": "window",
                    "module": "timeflux.nodes.window",
                    "class": "Window",
                    "params": {"length": window_length, "step": window_step},
                },
                {
                    "id": "criteria",
                    "module": "timeflux.nodes.apply",
                    "class": "ApplyMethod",
                    "params": {
                        "method": f"numpy.{method}",
                        "apply_mode": "reduce",
                    },
                },
                {
                    "id": "discretize",
                    "module": "timeflux_dsp.nodes.quality",
                    "class": "Discretize",
                    "params": {"range": range},
                },
            ],
            "edges": [
                {"source": "window", "target": "criteria"},
                {"source": "criteria", "target": "discretize"},
            ],
        }
        self.load(graph)

    def update(self):
        # copy the meta
        self.o.meta = self.i.meta

        # When we have not received data, there is nothing to do
        if self.i.data is None or self.i.data.empty:
            return

        self.set_port("window", port_id="i", data=self.i.data, meta=self.i.meta)

        self.run()

        self.o = self.get_port("discretize", port_id="o")



"""This module contains nodes for spectral analysis with Timeflux."""

import numpy as np
import pandas as pd
import xarray as xr
from scipy.signal import welch
from scipy.fft import fftfreq, rfftfreq, fft, rfft

from timeflux.core.node import Node


class FFT(Node):
    """Compute the one-dimensional discrete Fourier Transform for each column using the Fast Fourier Tranform algorithm.

    Attributes:
        i (Port): default input, expects DataFrame.
        o (Port): default output, provides DataArray.

    Example:

        In this exemple, we simulate a white noise and we apply FFT:

         * ``fs`` = `10.0`
         * ``nfft`` = `5`
         * ``return_onesided`` = `False`

        self.i.data::

                                              A         B         C
            2017-12-31 23:59:59.998745401  0.185133  0.541901  0.872946
            2018-01-01 00:00:00.104507143  0.732225  0.806561  0.658783
            2018-01-01 00:00:00.202319939  0.692277  0.849196  0.249668
            2018-01-01 00:00:00.300986584  0.489425  0.221209  0.987668
            2018-01-01 00:00:00.396560186  0.944059  0.039427  0.705575


        self.o.data::

            xarray.DataArray (times: 1, freqs: 5, space: 3)
            array([[[ 3.043119+0.j      ,  2.458294+0.j      ,  3.47464 +0.j      ],
                    [-0.252884+0.082233j, -0.06265 -1.098709j,  0.29353 +0.478287j],
                    [-0.805843+0.317437j,  0.188256+0.146341j,  0.151515-0.674376j],
                    [-0.805843-0.317437j,  0.188256-0.146341j,  0.151515+0.674376j],
                    [-0.252884-0.082233j, -0.06265 +1.098709j,  0.29353 -0.478287j]]])
            Coordinates:
              * times    (times) datetime64[ns] 2018-01-01T00:00:00.396560186
              * freqs    (freqs) float64 0.0 2.0 4.0 -4.0 -2.0
              * space    (space) object 'A' 'B' 'C'

    Notes:
       This node should be used after a buffer.

    References:
        * `scipy.fft <https://docs.scipy.org/doc/scipy/reference/fft.html>`_

    """

    def __init__(self, fs=1.0, nfft=None, return_onesided=True):
        """
        Args:
            fs (float): Nominal sampling rate of the input data.
            nfft (int|None): Length of the Fourier transform. Default: length of the chunk.
            return_onesided (bool): If `True`, return a one-sided spectrum for real data.
                                          If `False` return a two-sided spectrum.
                                          (Note that for complex data, a two-sided spectrum is always returned.)
                                          Default: `True`.
        """

        self._fs = fs
        self._nfft = nfft
        if return_onesided:
            self._sides = "onesided"
        else:
            self._sides = "twosided"
        if self._nfft is not None:
            self._set_freqs()

    def _check_nfft(self):

        # Check validity of nfft at first chunk
        if self._nfft is None:
            self.logger.debug("nfft := length of the chunk ")
            self._nfft = self.i.data.shape[0]
            self._set_freqs()
        elif self._nfft < self.i.data.shape[0]:
            raise ValueError("nfft must be greater than or equal to length of chunk.")
        else:
            self._nfft = int(self._nfft)

    def _set_freqs(self):

        # Set freqs indexes
        if self._sides == "onesided":
            self._freqs = rfftfreq(self._nfft, 1 / self._fs)
        else:
            self._freqs = fftfreq(self._nfft, 1 / self._fs)

    def update(self):

        # copy the meta
        self.o = self.i

        # When we have not received data, there is nothing to do
        if not self.i.ready():
            return

        # At this point, we are sure that we have some data to process
        self._check_nfft()
        self.o.data = self.i.data
        if self._sides == "twosided":
            func = fft
        else:
            self.o.data = self.o.data.apply(lambda x: x.real)
            func = rfft
        values = func(self.o.data.values.T, n=self._nfft).T
        self.o.data = xr.DataArray(
            np.stack([values], 0),
            coords=[[self.o.data.index[-1]], self._freqs, self.o.data.columns],
            dims=["time", "freq", "space"],
        )


class Welch(Node):
    """Estimate power spectral density using Welch’s method.

    Attributes:
       i (Port): default input, expects DataFrame.
       o (Port): default output, provides DataArray with dimensions (time, freq, space).

    Example:

    In this exemple, we simulate data with noisy sinus on three sensors (columns `a`, `b`, `c`):

        * ``fs`` = `100.0`
        * ``nfft`` = `24`

    node.i.data::
        \s                       a         b         c
        1970-01-01 00:00:00.000 -0.233920 -0.343296  0.157988
        1970-01-01 00:00:00.010  0.460353  0.777296  0.957201
        1970-01-01 00:00:00.020  0.768459  1.234923  1.942190
        1970-01-01 00:00:00.030  1.255393  1.782445  2.326175
        ...                      ...       ...       ...
        1970-01-01 00:00:01.190  1.185759  2.603828  3.315607

    node.o.data::

        <xarray.DataArray (time: 1, freq: 13, space: 3)>
        array([[[2.823924e-02, 1.087382e-01, 1.153163e-01],
            [1.703466e-01, 6.048703e-01, 6.310628e-01],
            ...            ...           ...
            [9.989429e-04, 8.519226e-04, 7.769918e-04],
            [1.239551e-03, 7.412518e-04, 9.863335e-04],
            [5.382880e-04, 4.999334e-04, 4.702757e-04]]])
        Coordinates:
            * time     (time) datetime64[ns] 1970-01-01T00:00:01.190000
            * freq     (freq) float64 0.0 4.167 8.333 12.5 16.67 ... 37.5 41.67 45.83 50.0
            * space    (space) object 'a' 'b' 'c'

    Notes:

        This node should be used after a Window with the appropriate length, with regard to the parameters
        `noverlap`, `nperseg` and `nfft`.
        It should be noted that a pipeline such as {LargeWindow-Welch} is in fact equivalent to a pipeline
        {SmallWindow-FFT-LargeWindow-Average} with SmallWindow 's parameters `length` and `step` respectively
        equivalent to `nperseg` and `step` and with FFT node with same kwargs.

    """

    def __init__(self, rate=None, closed="right", **kwargs):
        """
        Args:
            rate (float|None): Nominal sampling rate of the input data. If `None`, the rate will be taken from the input meta/
            closed (str): Make the index closed on the `right`, `left` or `center`.
            kwargs:  Keyword arguments to pass to scipy.signal.welch function.
                            You can specify: window, nperseg, noverlap, nfft, detrend, return_onesided and scaling.
        """

        self._rate = rate
        self._closed = closed
        self._kwargs = kwargs
        self._set_default()

    def _set_default(self):
        # We set the default params if they are not specifies in kwargs in order to check that they are valid, in respect of the length and sampling of the input data.
        if "nperseg" not in self._kwargs.keys():
            self._kwargs["nperseg"] = 256
            self.logger.debug("nperseg := 256")
        if "nfft" not in self._kwargs.keys():
            self._kwargs["nfft"] = self._kwargs["nperseg"]
            self.logger.debug(
                "nfft := nperseg := {nperseg}".format(nperseg=self._kwargs["nperseg"])
            )
        if "noverlap" not in self._kwargs.keys():
            self._kwargs["noverlap"] = self._kwargs["nperseg"] // 2
            self.logger.debug(
                "noverlap := nperseg/2 := {noverlap}".format(
                    noverlap=self._kwargs["noverlap"]
                )
            )

    def _check_nfft(self):
        # Check validity of nfft at first chun
        if not all(
            i <= len(self.i.data)
            for i in [self._kwargs[k] for k in ["nfft", "nperseg", "noverlap"]]
        ):
            raise ValueError(
                "nfft, noverlap and nperseg must be greater than or equal to length of chunk."
            )
        else:
            self._kwargs.update(
                {
                    keyword: int(self._kwargs[keyword])
                    for keyword in ["nfft", "nperseg", "noverlap"]
                }
            )

    def update(self):
        # copy the meta
        self.o = self.i

        # When we have not received data, there is nothing to do
        if not self.i.ready():
            return

        # Check rate
        if self._rate:
            rate = self._rate
        elif "rate" in self.i.meta:
            rate = self.i.meta["rate"]
        else:
            raise ValueError(
                "The rate was neither explicitely defined nor found in the stream meta."
            )

        # At this point, we are sure that we have some data to process
        # apply welch on the data:
        self._check_nfft()
        f, Pxx = welch(x=self.i.data, fs=rate, **self._kwargs, axis=0)

        if self._closed == "left":
            time = self.i.data.index[-1]
        elif self._closed == "center":

            def middle(a):
                return int(np.ceil(len(a) / 2)) - 1

            time = self.i.data.index[middle(self.i.data)]
        else:  # right
            time = self.i.data.index[-1]
        # f is the frequency axis and Pxx the average power of shape (Nfreqs x Nchanels)
        # we reshape Pxx to fit the ('time' x 'freq' x 'space') dimensions
        self.o.data = xr.DataArray(
            np.stack([Pxx], 0),
            coords=[[time], f, self.i.data.columns],
            dims=["time", "frequency", "space"],
        )


class Bands(Node):
    """Averages the XArray values over freq dimension according to the frequencies bands given in arguments.

    This node selects a subset of values over the chosen dimensions, averages them along this axis and convert the result into a flat dataframe.
    This node will output as many ports bands as given bands, with their respective name as suffix.

        Attributes:
            i (Port): default output, provides DataArray with 3 dimensions (time, freq, space).
            o (Port): Default output, provides DataFrame.
            o_* (Port): Dynamic outputs, provide DataFrame.

    """

    def __init__(self, bands=None, relative=False):

        """
        Args:
           bands (dict): Define the band to extract given its name and its range.
                         An output port will be created with the given names as suffix.

        """
        bands = bands or {
            "delta": [1, 4],
            "theta": [4, 8],
            "alpha": [8, 12],
            "beta": [12, 30],
        }
        self._relative = relative
        self._bands = []
        for band_name, band_range in bands.items():
            self._bands.append(
                dict(
                    port=getattr(self, "o_" + band_name),
                    slice=slice(band_range[0], band_range[1]),
                    meta={"bands": {"range": band_range, "relative": relative}},
                )
            )

    def update(self):

        # When we have not received data, there is nothing to do
        if not self.i.ready():
            return

        # At this point, we are sure that we have some data to process
        for band in self._bands:
            # 1. select the Xarray on freq axis in the range, 2. average along freq axis
            band_power = (
                self.i.data.loc[{"frequency": band["slice"]}].sum("frequency").values
            )  # todo: sum
            if self._relative:
                tot_power = self.i.data.sum("frequency").values
                tot_power[tot_power == 0.0] = 1
                band_power /= tot_power

            band["port"].data = pd.DataFrame(
                columns=self.i.data.space.values,
                index=self.i.data.time.values,
                data=band_power,
            )
            band["port"].meta = {**(self.i.meta or {}), **band["meta"]}



import pandas as pd

from timeflux.core.node import Node


class Concat(Node):
    """Concat list of data ports .
    Attributes:
        i_* (Port): Dynamic inputs, expects DataFrame and meta.
        o (Port): Default output, provides DataFrame.
    Args:
        axis (str|int): The axis to concatenate along.
        kwargs: Keyword arguments to pass to pd.concat function

    Notes
    -----
    There is no shape verification in this node, that won't result in any Exception
    but may introduce NaN in the DataFrame. The responsability is left to the user.

    """

    def __init__(self, axis=1, **kwargs):
        self._axis = axis
        self._kwargs = kwargs

    def update(self):
        ports = list(self.iterate("i*"))
        i_data = [port.data for (name, _, port) in ports if port.data is not None]
        if i_data:
            self.o.data = pd.concat(i_data, axis=self._axis, **self._kwargs)
